#include "LTR22AdcConfig.h"
#include "LTR22.h"
#include "LTR22TypeInfo.h"
#include "ltr/include/ltr22api.h"
#include "lqmeas/EnumConfigKey.h"
#include "lqmeas/StdConfigKeys.h"
#include <QJsonObject>
#include <QJsonArray>

namespace LQMeas {
    static const QLatin1String cfgkey_ch_enabled         {StdConfigKeys::enabled()};
    static const QLatin1String cfgkey_ch_range_num       {StdConfigKeys::rangeNum()};
    static const QLatin1String cfgkey_channels_arr       {StdConfigKeys::channels()};
    static const QLatin1String cfgkey_ac_mode            {StdConfigKeys::acMode()};
    static const QLatin1String cfgkey_adc_freq_code      {StdConfigKeys::adcFreqCode()};
    static const QLatin1String cfgkey_adc_zero_mode      {"ZeroMode"};

    typedef EnumConfigKey<LTR22AdcConfig::AcMode> LTR22AcModeKey;
    static const LTR22AcModeKey adc_acmodes[] = {
        {QStringLiteral("acdc"),        LTR22AdcConfig::AcMode::AcDc},
        {QStringLiteral("ac"),          LTR22AdcConfig::AcMode::AcOnly},
        {LTR22AcModeKey::defaultKey(),  LTR22AdcConfig::AcMode::AcDc},
    };


    LTR22AdcConfig::LTR22AdcConfig(const DevAdcInfo &info) : DevAdcConfig{info} {
        memset(&params, 0, sizeof(params));
        for (int ch=0; ch < adc_channels_cnt; ++ch) {
            params.Ch[ch].enabled = false;
            params.Ch[ch].range_num = defaultRangeNum();
        }
    }

    LTR22AdcConfig::LTR22AdcConfig(const LTR22AdcConfig &cfg) : DevAdcConfig{cfg} {
         memcpy(&params, &cfg.params, sizeof (params));
    }

    double LTR22AdcConfig::adcFreq() const {
        return LTR22TypeInfo::typeAdcFreqVal(params.freq_num);
    }

    double LTR22AdcConfig::typeAdcFreqVal(LTR22AdcConfig::AdcFreq freq) {
        return LTR22TypeInfo::typeAdcFreqVal(typeAdcFreqNum(freq));
    }

    void LTR22AdcConfig::adcSetChEnabled(int ch, bool en) {
        if (params.Ch[ch].enabled != en) {
            params.Ch[ch].enabled = en;
            notifyChEnableStateChanged(ch);
            notifyConfigChanged();
        }
    }

    void LTR22AdcConfig::adcSetChRangeNum(int ch, int range_num) {
        if (params.Ch[ch].range_num != range_num) {
            params.Ch[ch].range_num = range_num;
            notifyChRangeChanged(ch);
            notifyConfigChanged();
        }
    }

    void LTR22AdcConfig::adcSetAcMode(LTR22AdcConfig::AcMode mode) {
        if (params.ac_mode != mode) {
            params.ac_mode = mode;
            notifyChModeChanged(-1);
            notifyConfigChanged();
        }
    }

    void LTR22AdcConfig::adcSetFreqNum(int freq_num) {
        if (params.freq_num != freq_num) {
            params.freq_num = freq_num;
            notifyChFreqChanged();
            notifyAdcFreqChanged();
            notifyConfigChanged();
        }
    }

    void LTR22AdcConfig::adcSetZeroModeEnabled(bool en) {
        if (params.zero_mode_en != en) {
            params.zero_mode_en = en;
            notifyConfigChanged();
        }
    }


    void LTR22AdcConfig::protSave(QJsonObject &cfgObj) const     {
        QJsonArray chArray;
        for (int ch=0; ch < adc_channels_cnt; ++ch) {
            QJsonObject chObj;
            chObj[cfgkey_ch_enabled] = params.Ch[ch].enabled;
            chObj[cfgkey_ch_range_num] = params.Ch[ch].range_num;
            chArray.append(chObj);
        }
        cfgObj[cfgkey_channels_arr] = chArray;
        cfgObj[cfgkey_ac_mode] =LTR22AcModeKey::getKey(adc_acmodes, params.ac_mode);
        cfgObj[cfgkey_adc_freq_code] = params.freq_num;
        cfgObj[cfgkey_adc_zero_mode] = params.zero_mode_en;
    }

    void LTR22AdcConfig::protLoad(const QJsonObject &cfgObj)     {
        const QJsonArray &chArray = cfgObj[cfgkey_channels_arr].toArray();
        int ch_cnt = chArray.size();

        for (int ch=0; (ch < ch_cnt) && (ch < adc_channels_cnt); ++ch) {
            const QJsonObject chObj = chArray.at(ch).toObject();
            params.Ch[ch].enabled = chObj[cfgkey_ch_enabled].toBool();
            int range = chObj[cfgkey_ch_range_num].toInt(defaultRangeNum());
            if ((range >= 0) || (range < adcInfo()->adcChRangesCnt(ch, adcUnitMode(ch)))) {
                params.Ch[ch].range_num = range;
            } else {
                params.Ch[ch].range_num = defaultRangeNum();
            }

        }

        for (int ch=ch_cnt; ch < adc_channels_cnt; ++ch) {
            params.Ch[ch].enabled = false;
            params.Ch[ch].range_num = defaultRangeNum();
        }

        params.ac_mode = LTR22AcModeKey::getValue(adc_acmodes, cfgObj[cfgkey_ac_mode].toString());
        params.freq_num = cfgObj[cfgkey_adc_freq_code].toInt(defaultFreqNum());
        params.zero_mode_en = cfgObj[cfgkey_adc_zero_mode].toBool(false);
    }
}
