#include "LTR27Info.h"
#include <QJsonObject>
#include <QJsonArray>
#include "lqmeas/StdConfigKeys.h"

namespace LQMeas{
    static const QLatin1String cfgkey_mezzanines_group {"Mezzanines"};

    LTR27Info::LTR27Info(const LTR27TypeInfo &type, const QString &serial, int mcu_ver) :
        DeviceInfo{type, serial},
        m_mcu_ver{mcu_ver} {

        for (int mez_idx = 0; mez_idx < LTR27TypeInfo::mezzanines_cnt; ++mez_idx) {
            m_mezzanines.append(nullptr);
        }
    }

    LTR27Info::LTR27Info(QList<LTR27MezzanineInfo *> mezzanines, const QString &serial, int mcu_ver) :
        DeviceInfo{LTR27TypeInfo{mezzaninesTypesList(mezzanines)}, serial},
        m_mcu_ver{mcu_ver} {

        for (int mez_idx = 0; mez_idx < LTR27TypeInfo::mezzanines_cnt; ++mez_idx) {
            m_mezzanines.append(mez_idx < mezzanines.size() ? mezzanines.at(mez_idx) : nullptr);
        }
    }

    LTR27Info::~LTR27Info() {
        qDeleteAll(m_mezzanines);
    }

    QString LTR27Info::mcuVerStr() const {
        return QString("%0.%1.%2.%3")
                .arg(QString::number((m_mcu_ver >> 24) & 0xFF))
                .arg(QString::number((m_mcu_ver >> 16) & 0xFF))
                .arg(QString::number((m_mcu_ver >>  8) & 0xFF))
                .arg(QString::number((m_mcu_ver >>  0) & 0xFF));
    }

    void LTR27Info::saveMezzanineInfo(QJsonObject &infoObj, const QList<LTR27MezzanineInfo *> &mezInfoList)  {
        QJsonArray mezArray;
        for (int mez_idx = 0; mez_idx < LTR27TypeInfo::mezzanines_cnt; ++mez_idx) {
            QJsonObject mezObj;
            LTR27MezzanineInfo *mezInfo = mez_idx < mezInfoList.size() ? mezInfoList.at(mez_idx) : nullptr;
            if (mezInfo && mezInfo->type()) {
                mezObj[StdConfigKeys::type()] = mezInfo->type()->idName();
                mezObj[StdConfigKeys::serialNumber()] = mezInfo->serial();
            }
            mezArray.append(mezObj);
        }
        infoObj[cfgkey_mezzanines_group] = mezArray;
    }

    void LTR27Info::protSave(QJsonObject &infoObj) const {
        infoObj[StdConfigKeys::mcuVer()] = m_mcu_ver;
        saveMezzanineInfo(infoObj, m_mezzanines);
    }

    void LTR27Info::protLoad(const QJsonObject &infoObj) {
        m_mcu_ver = infoObj[StdConfigKeys::mcuVer()].toInt();
        const QJsonArray mezArray =  infoObj[cfgkey_mezzanines_group].toArray();
        qDeleteAll(m_mezzanines);
        for (int mez_idx = 0; mez_idx < LTR27TypeInfo::mezzanines_cnt; ++mez_idx) {
            LTR27MezzanineInfo *info = nullptr;
            if (mez_idx < mezArray.size()) {
                QJsonObject mezObj = mezArray.at(mez_idx).toObject();
                QString mezTypeId =  mezObj[StdConfigKeys::type()].toString();
                QList<const LTR27MezzanineTypeInfo *> mezTypes = LTR27MezzanineTypeInfo::allTypes();
                Q_FOREACH(const LTR27MezzanineTypeInfo *mezType, mezTypes) {
                    if (mezType->idName() == mezTypeId) {
                        info = new LTR27MezzanineInfo{mezType, mezObj[StdConfigKeys::serialNumber()].toString()};
                    }
                }
            }
            m_mezzanines[mez_idx] = info;
        }
        setType(new LTR27TypeInfo{mezzaninesTypesList(m_mezzanines)});
    }

    QList<const LTR27MezzanineTypeInfo *> LTR27Info::mezzaninesTypesList(
            const QList<LTR27MezzanineInfo *> &infoList) {
        QList<const LTR27MezzanineTypeInfo *> ret;
        for (LTR27MezzanineInfo *info : infoList) {
            ret.append(info ? info->type() : nullptr);
        }
        return ret;
    }

    LTR27Info::LTR27Info(const LTR27Info &info): DeviceInfo{info}, m_mcu_ver{info.m_mcu_ver} {
        for (int mez_idx = 0; mez_idx < LTR27TypeInfo::mezzanines_cnt; ++mez_idx) {
            m_mezzanines.append(info.m_mezzanines.at(mez_idx) ?
                                new LTR27MezzanineInfo{*info.m_mezzanines.at(mez_idx)} : nullptr);
        }
    }
}
