#include "LTR43.h"
#include "LTR43Info.h"
#include "LTR43Config.h"
#include "lqmeas/lqtdefs.h"
#include "lqmeas/StdErrors.h"
#include "lqmeas/devs/LTR/service/LTRServiceRef.h"
#include "lqmeas/devs/LTR/crates/LTRCrate.h"
#include "lqmeas/devs/DeviceRef.h"

namespace LQMeas {
    LTR43::LTR43(LTRCrate *crate, int slot, QObject *parent) :
        LTRModule{crate, slot, typeModuleID, new LTR43Config{}, LTR43Info{}, parent} {

        LTR43_Init(&m_hnd);
    }

    LTR43::~LTR43() {
        LTR43_Close(&m_hnd);
    }

    bool LTR43::isOpened() const {
        QMutexLocker lock(&m_io_mutex);
        return LTR43_IsOpened(&m_hnd) == LTR_OK;
    }

    QString LTR43::errorString(int err) const {
        return QSTRING_FROM_CSTR(LTR43_GetErrorString(err));
    }

    const LTR43Config &LTR43::devspecConfig() const {
        return static_cast<const LTR43Config &>(config());
    }

    const LTR43Info &LTR43::devspecInfo() const {
        return static_cast<const LTR43Info &>(devInfo());
    }

    TLTR *LTR43::channel() const {
        return &m_hnd.Channel;
    }

    void LTR43::protOpen(OpenFlags flags, LQError &err) {
        Q_UNUSED(flags)
        QMutexLocker lock{&m_io_mutex};
        INT ltr_err = LTR43_Open(&m_hnd,
                                 serviceRef().ltrdIPv4Addr(),
                                 serviceRef().ltrdTcpPort(),
                                 ltrCrateRef().devInfo().serial().toLatin1(),
                                 slot());

        if (ltr_err == LTR_OK) {
            setDeviceInfo(LTR43Info(LTR43TypeInfo::defaultInfo(),
                                   QSTRING_FROM_CSTR(m_hnd.ModuleInfo.Serial),
                                   QSTRING_FROM_CSTR(m_hnd.ModuleInfo.FirmwareVersion)));
        }

        if (ltr_err != LTR_OK) {
            LTR43_Close(&m_hnd);
            getError(ltr_err, err);
        }
    }

    void LTR43::protClose(LQError &err) {
        QMutexLocker lock{&m_io_mutex};
        getError(LTR43_Close(&m_hnd), err);
    }

    void LTR43::protConfigure(const DeviceConfig &cfg, LQError &err) {
        QMutexLocker lock{&m_io_mutex};
        const LTR43Config *setCfg {qobject_cast<const LTR43Config *>(&cfg)};
        if (setCfg) {
            m_hnd.IO_Ports.Port1 = portCfgWord(setCfg, 0);
            m_hnd.IO_Ports.Port2 = portCfgWord(setCfg, 1);
            m_hnd.IO_Ports.Port3 = portCfgWord(setCfg, 2);
            m_hnd.IO_Ports.Port4 = portCfgWord(setCfg, 3);

            getError(LTR43_Config(&m_hnd), err);
        } else {
            err = StdErrors::InvalidConfigType();
        }
    }

    void LTR43::protInAsyncDig(unsigned &val, LQError &err) {
        QMutexLocker lock{&m_io_mutex};
        DWORD rd_data;
        INT err_code = LTR43_ReadPort(&m_hnd, &rd_data);
        if (err_code == LTR_OK) {
            val = rd_data;
        } else {
            getError(err_code, err);
        }
    }

    void LTR43::protOutAsyncDigAll(unsigned val, LQError &err) {
        QMutexLocker lock{&m_io_mutex};
        getError(LTR43_WritePort(&m_hnd, val), err);
    }

    int LTR43::portCfgWord(const LTR43Config *cfg, int port_idx) {
        return cfg->portDirection(port_idx) == LTR43Config::Direction::Out ?
                    LTR43_PORT_DIR_OUT : LTR43_PORT_DIR_IN;
    }
}
