#include "LTRT13.h"
#include "LTRT13Config.h"
#include "LTRT13Info.h"
#include "lqmeas/lqtdefs.h"
#include "lqmeas/StdErrors.h"
#include "lqmeas/devs/LTR/service/LTRServiceRef.h"
#include <QDateTime>

extern "C" LTRT13API_DllExport(INT) LTRT13_WriteInfo(TLTRT13 *hnd);

namespace LQMeas {
    LTRT13::~LTRT13() {
        LTRT13_Close(&m_hnd);
    }

    bool LTRT13::isOpened() const {
        return LTRT13_IsOpened(&m_hnd) == LTR_OK;
    }

    QString LTRT13::errorString(int err) const {
        return QSTRING_FROM_CSTR(LTRT13_GetErrorString(err));;
    }

    const LTRT13Config &LTRT13::devspecConfig() const {
        return static_cast<const LTRT13Config &>(config());
    }


    const LTRT13Info &LTRT13::devspecInfo() const {
        return static_cast<const LTRT13Info &>(devInfo());
    }

    void LTRT13::setExtStartLevel(bool lvl, LQError &err) {
        getError(LTRT13_SetExtStartLevel(&m_hnd, lvl), err);
    }

    void LTRT13::setExtSyncGenEnabled(bool en, LQError &err) {
        getError(LTRT13_SetExtSyncGenEnabled(&m_hnd, en), err);
    }

    void LTRT13::setMarksGenEnabled(bool en, LQError &err) {
        getError(LTRT13_SetMarksGenEnabled(&m_hnd, en), err);
    }

    bool LTRT13::extStartLevel() const {
        return isOpened() && m_hnd.State.ExtStartLevel;
    }

    bool LTRT13::extSyncGenIsEnabled() const {
        return isOpened() && m_hnd.State.ExtSyncGenEn;
    }

    bool LTRT13::markGenIsEnabled() const {
        return isOpened() && m_hnd.State.MarkGenEn;
    }

    double LTRT13::rShuntCbrValue() const {
        return m_hnd.ModuleInfo.RValue;
    }

    QDateTime LTRT13::cbrTime() const {
        return m_hnd.ModuleInfo.CbrTime == 0 ? QDateTime() : QDateTime::fromMSecsSinceEpoch(m_hnd.ModuleInfo.CbrTime * 1000);
    }

    double LTRT13::rShuntNomValue() {
        return LTRT13_RVAL_NOM;
    }

    void LTRT13::writeCbrValue(double rval, LQError &err) {
        m_hnd.ModuleInfo.RValue = rval;
        m_hnd.ModuleInfo.CbrTime = QDateTime::currentMSecsSinceEpoch();
        getError(LTRT13_WriteInfo(&m_hnd), err);
    }

    void LTRT13::eraseCbrValue(LQError &err) {
        m_hnd.ModuleInfo.RValue = LTRT13_RVAL_NOM;
        m_hnd.ModuleInfo.CbrTime = 0;
        getError(LTRT13_WriteInfo(&m_hnd), err);
    }

    TLTR *LTRT13::channel() const {
        return &m_hnd.Channel;
    }

    void LTRT13::protOpen(Device::OpenFlags flags, LQError &err) {
        INT ltr_err = LTRT13_Open(&m_hnd,
                                 serviceRef().ltrdIPv4Addr(),
                                 serviceRef().ltrdTcpPort(),
                                 ltrCrateRef().devInfo().serial().toLatin1(),
                                 static_cast<WORD>(slot()));
        if (LTRT13_IsOpened(&m_hnd) == LTR_OK) {
            setDeviceInfo(LTRT13Info{LTRT13TypeInfo::defaultInfo(),
                                     QSTRING_FROM_CSTR(m_hnd.ModuleInfo.Serial),
                                     m_hnd.ModuleInfo.VerPLD});
        }

        if (ltr_err != LTR_OK) {
            if (!(flags & OpenFlag::Force) || (ltr_err == LTR_WARNING_MODULE_IN_USE))
                LTRT13_Close(&m_hnd);
            getError(ltr_err, err);
        }
    }

    void LTRT13::protClose(LQError &err) {
        getError(LTRT13_Close(&m_hnd), err);
    }

    void LTRT13::protConfigure(const DeviceConfig &cfg, LQError &err) {
        const LTRT13Config *setCfg {qobject_cast<const LTRT13Config *>(&cfg)};
        if (setCfg) {
            INT ltr_err {LTRT13_FillExtSyncFreqParams(&m_hnd, setCfg->extSyncFreq(), NULL)};
            if ((ltr_err == LTR_OK) && extSyncGenIsEnabled()) {
                /* повторное разрешение генерации для обновления частоты */
                ltr_err = LTRT13_SetExtSyncGenEnabled(&m_hnd, TRUE);
            }
            getError(ltr_err, err);
        }  else {
            err += StdErrors::InvalidConfigType();
        }
    }

    LTRT13::LTRT13(LQMeas::LTRCrate *crate, int slot, QObject *parent) :
        LTR01(crate, slot, typeModuleSubID, new LTRT13Config{}, LTRT13Info{}, parent) {

        LTRT13_Init(&m_hnd);
    }
}
