#include "E502DeviceRefIPAddr.h"
#include "E502.h"
#include "e502api.h"
#include "lqmeas/devs/DeviceInterface.h"
#include "lqmeas/StdConfigKeys.h"
#include <QJsonObject>

static const QLatin1String cfgkey_ip_group {"IP"};
static const QLatin1String cfgkey_cmd_port {"CmdPort"};
static const QLatin1String cfgkey_data_port {"DataPort"};
static const QLatin1String cfgkey_custom_port_group {"CustomPorts"};

namespace LQMeas {

    E502DeviceRefIPAddr::E502DeviceRefIPAddr(const DeviceInfo &devinfo,
                                             const QHostAddress &ipAddr, int conTout) :
        DeviceRef{devinfo, DeviceInterfaces::ethernet()},
        m_ipAddr{ipAddr},
        m_conTout{conTout},
        m_customPorts{false},
        m_cmdPort{defaultCmdPort()},
        m_dataPort{defaultDataPort()} {

    }

    E502DeviceRefIPAddr::E502DeviceRefIPAddr(const E502DeviceRefIPAddr &ipRef) :
        DeviceRef{ipRef},
        m_ipAddr{ipRef.m_ipAddr},
        m_conTout{ipRef.m_conTout},
        m_customPorts{ipRef.m_customPorts},
        m_cmdPort{ipRef.m_cmdPort},
        m_dataPort{ipRef.dataPort()} {

    }

    E502DeviceRefIPAddr::E502DeviceRefIPAddr(const DeviceInfo &devinfo,
                                             const E502DeviceRefIPAddr &ipRef) :
        DeviceRef{devinfo, DeviceInterfaces::ethernet()},
        m_ipAddr{ipRef.m_ipAddr},
        m_conTout{ipRef.m_conTout},
        m_customPorts{ipRef.m_customPorts},
        m_cmdPort{ipRef.m_cmdPort},
        m_dataPort{ipRef.dataPort()} {

    }

    const QString &E502DeviceRefIPAddr::typeRefTypeID() {
        static const QString str { QStringLiteral("e502ipaddr") };
        return str;
    }

    int E502DeviceRefIPAddr::defaultConnectionTimeout() {
        return E502::defaultEthConnectionTout();
    }

    DeviceRef::DevCheckResult E502DeviceRefIPAddr::checkDevice(const Device &device) const {
        DevCheckResult res = DevCheckResult::BadDevice;
        const E502 *e502 = qobject_cast<const E502 *>(&device);
        if (e502 && e502->ipAddrRef()) {
            if (e502->ipAddrRef()->sameDevice(*this)) {
                res = DevCheckResult::GoodDevice;
            }
        }
        return res;
    }

    bool E502DeviceRefIPAddr::sameDevice(const DeviceRef &ref) const  {
        bool same = ref.refTypeID() == typeRefTypeID();
        if (same) {
            const E502DeviceRefIPAddr &ipRef = static_cast<const E502DeviceRefIPAddr &>(ref);
            same = (ipRef.ipAddr() == ipAddr())
                    && (ipRef.cmdPort() == cmdPort())
                    && (ipRef.dataPort() == dataPort());
        }
        return  same;
    }

    void E502DeviceRefIPAddr::save(QJsonObject &refObj) const {
        DeviceRef::save(refObj);

        QJsonObject ipObj;
        ipObj[StdConfigKeys::ipv4Addr()] = m_ipAddr.toString();
        ipObj[StdConfigKeys::conTout()] = m_conTout;
        QJsonObject portObj;
        portObj[StdConfigKeys::enabled()] = m_customPorts;
        portObj[cfgkey_cmd_port] = m_cmdPort;
        portObj[cfgkey_data_port] = m_dataPort;
        ipObj[cfgkey_custom_port_group] = portObj;

        refObj[cfgkey_ip_group] = ipObj;
    }

    void E502DeviceRefIPAddr::load(const QJsonObject &refObj) {
        DeviceRef::load(refObj);

        const QJsonObject &ipObj = refObj[cfgkey_ip_group].toObject();
        m_ipAddr = QHostAddress{ipObj[StdConfigKeys::ipv4Addr()].toString()};
        m_conTout = ipObj[StdConfigKeys::conTout()].toInt(defaultConnectionTimeout());

        const QJsonObject &portObj = ipObj[cfgkey_custom_port_group].toObject();
        m_customPorts = portObj[StdConfigKeys::enabled()].toBool(false);
        m_cmdPort     = portObj[cfgkey_cmd_port].toInt(defaultCmdPort());
        m_dataPort    = portObj[cfgkey_data_port].toInt(defaultDataPort());
    }

    DeviceRef *E502DeviceRefIPAddr::clone() const {
        return new E502DeviceRefIPAddr(*this);
    }

    QString E502DeviceRefIPAddr::displayName() const {
        QString ret = QString("%1, %2 %3").arg(DeviceRef::displayName())
                .arg(E502::tr("IP"))
                .arg(m_ipAddr.toString());
        if (m_customPorts) {
            ret.append(QString(":%1/%2").arg(m_cmdPort).arg(m_dataPort));
        }
        return  ret;
    }

    QString E502DeviceRefIPAddr::protKey() const {
        QString ret = QString("%1_ip%2").arg(DeviceRef::protKey()).arg(m_ipAddr.toString());
        if (m_customPorts) {
            ret.append(QString("_p%1_%2").arg(m_cmdPort).arg(m_dataPort));
        }
        return ret;
    }

    t_x502_devrec E502DeviceRefIPAddr::createDevrec() const {
        t_x502_devrec devrec;
        int32_t err = E502_MakeDevRecordByIpAddr(&devrec, m_ipAddr.toIPv4Address(),
                                                 0, static_cast<uint32_t>(m_conTout));
        if (err == X502_ERR_OK) {
            if (m_customPorts) {
                E502_EthDevRecordSetCmdPort(&devrec, m_cmdPort & 0xFFFF);
                E502_EthDevRecordSetDataPort(&devrec, m_dataPort & 0xFFFF);
            }
        }
        return  devrec;

    }
}
