#include "E502DeviceRefNetSvc.h"
#include "lqmeas/devs/DeviceInfo.h"
#include "lqmeas/devs/DeviceInterface.h"
#include "lqmeas/StdConfigKeys.h"
#include "../E502.h"
#include "e502api.h"
#include "E502NetSvcRecord.h"
#include <QJsonObject>

namespace LQMeas {
    static const QLatin1String cfgkey_net_svc_group {"NetSvc"};
    static const QLatin1String cfgkey_instance_name {"InstanceName"};

    E502DeviceRefNetSvc::E502DeviceRefNetSvc(const DeviceInfo &devinfo, const QString &instanceName, int conTout) :
        DeviceRef{devinfo, DeviceInterfaces::ethernet()},
        m_instanceName{instanceName},
        m_conTout{conTout} {

    }

    E502DeviceRefNetSvc::E502DeviceRefNetSvc(const E502DeviceRefNetSvc &ref) :
        DeviceRef{ref},
        m_instanceName{ref.m_instanceName},
        m_conTout{ref.m_conTout} {

    }

    E502DeviceRefNetSvc::E502DeviceRefNetSvc(const DeviceInfo &devinfo, const E502DeviceRefNetSvc &ref) :
        DeviceRef{devinfo, DeviceInterfaces::ethernet()},
        m_instanceName{ref.m_instanceName},
        m_conTout{ref.m_conTout} {

    }

    const QString &E502DeviceRefNetSvc::typeRefTypeID() {
        static const QString str { QStringLiteral("e502ethsvc") };
        return str;
    }

    int E502DeviceRefNetSvc::defaultConnectionTimeout() {
        return E502::defaultEthConnectionTout();
    }

    bool E502DeviceRefNetSvc::checkNetSvcRecord(const E502NetSvcRecord &rec) const  {
        QString serial {devInfo().serial()};
        return (rec.instanceName() == m_instanceName) &&
                   (serial.isEmpty() || (serial == rec.deviceSerial()));
    }

    LQMeas::DeviceRef::DevCheckResult LQMeas::E502DeviceRefNetSvc::checkDevice(const LQMeas::Device &device) const {
        DevCheckResult res {DevCheckResult::BadDevice};
        const E502 *e502 {qobject_cast<const E502 *>(&device)};
        if (e502 && e502->netRecord()) {
            QSharedPointer<const E502NetSvcRecord> rec = e502->netRecord();
            if (checkNetSvcRecord(*rec)) {
                res = DevCheckResult::GoodDevice;
            }
        }
        return res;
    }

    bool E502DeviceRefNetSvc::sameDevice(const DeviceRef &ref) const {
        bool same {ref.refTypeID() == typeRefTypeID()};
        if (same) {
            const E502DeviceRefNetSvc &svcRef = static_cast<const E502DeviceRefNetSvc &>(ref);
            same = svcRef.instanceName() == m_instanceName;
        }
        return  same;
    }

    void E502DeviceRefNetSvc::save(QJsonObject &refObj) const {
        DeviceRef::save(refObj);

        QJsonObject svcObj;
        svcObj[cfgkey_instance_name] = m_instanceName;
        svcObj[StdConfigKeys::conTout()] = m_conTout;
        refObj[cfgkey_net_svc_group] = svcObj;
    }

    void E502DeviceRefNetSvc::load(const QJsonObject &refObj) {
        DeviceRef::load(refObj);

        const QJsonObject &svcObj = refObj[cfgkey_net_svc_group].toObject();
        m_instanceName = svcObj[cfgkey_instance_name].toString();
        m_conTout = svcObj[StdConfigKeys::conTout()].toInt();
    }

    DeviceRef *E502DeviceRefNetSvc::clone() const {
        return new E502DeviceRefNetSvc(*this);
    }

    QString E502DeviceRefNetSvc::displayName() const {
        return QString{"%1, %2 '%3'"}.arg(DeviceRef::displayName(),
                                          E502::tr("Net Name"),
                                          m_instanceName);
    }

    QString E502DeviceRefNetSvc::protKey() const {
        return QString{"%1_netsvc%2"}.arg(DeviceRef::protKey(), m_instanceName);
    }

    t_x502_devrec E502DeviceRefNetSvc::createDevrec(const E502NetSvcRecord &rec) const     {
        return rec.createDevrec(m_conTout);
    }
}
