#include "x502.h"
#include "x502Info.h"
#include "x502Config.h"
#include "x502NativeError.h"
#include "lqmeas/lqtdefs.h"
#include "lqmeas/StdErrors.h"
#include "lqmeas/log/Log.h"
#include "lqmeas/devs/DeviceInterface.h"
#include "lqmeas/devs/DeviceRef.h"
#include "lqmeas/ifaces/out/SyncModes/stream/DevOutSyncModeImplStream.h"
#include "lqmeas/ifaces/out/SyncModes/cycle/DevOutSyncModeImplCycle.h"
#include <QMutexLocker>
#include <QStringBuilder>


namespace LQMeas {
    static t_x502_sync_mode f_conv_sync_mode(x502Config::SyncMode syncMode) {
        t_x502_sync_mode ret = X502_SYNC_INTERNAL;
        if (syncMode == x502Config::SyncMode::Internal) {
            ret = X502_SYNC_INTERNAL;
        } else if (syncMode == x502Config::SyncMode::ExternalMaster) {
            ret = X502_SYNC_EXTERNAL_MASTER;
        } else if (syncMode == x502Config::SyncMode::SYN1Rise) {
            ret = X502_SYNC_DI_SYN1_RISE;
        } else if (syncMode == x502Config::SyncMode::SYN1Fall) {
            ret = X502_SYNC_DI_SYN1_FALL;
        } else if (syncMode == x502Config::SyncMode::SYN2Rise) {
            ret = X502_SYNC_DI_SYN2_RISE;
        } else if (syncMode == x502Config::SyncMode::SYN2Fall) {
            ret = X502_SYNC_DI_SYN2_FALL;
        }
        return ret;
    }

    static t_x502_ref_freq f_conv_ref_freq(x502Config::RefFreq refFreq) {
        return  refFreq == x502Config::RefFreq::Freq_1500KHz ? X502_REF_FREQ_1500KHZ : X502_REF_FREQ_2000KHZ;
    }


    x502::x502(const t_x502_devrec &devrec, const x502Info &info) :
        Device{new x502Config{}, info, devrecIface(devrec)},
        DevAdc{this, info.devspecTypeInfo().adcChannelsCnt()},
        DevOutSync{this},
        DevOutAsyncDac{this},
        DevOutDacConverter{this},
        m_receiver{this},
        m_sender{this} {

        m_devrec = devrec;
        m_hnd = X502_Create();

        outSyncAddMode(new DevOutSyncModeImplStream{this, this});
        outSyncAddMode(new DevOutSyncModeImplCycle{this, this});
    }

    void x502::setNewDevRecord(const t_x502_devrec &devrec) {
        X502_FreeDevRecordList(&m_devrec,1);
        m_devrec = devrec;
    }


    x502::~x502() {
        X502_Free(m_hnd);
        X502_FreeDevRecordList(&m_devrec,1);
    }

    QString x502::location() const {
        return QString::fromUtf8(m_devrec.location);
    }

    bool x502::isOpened() const {
        return m_opened;
    }

    bool x502::checkConnection() const {
        bool ok = isOpened();
        if (ok && (X502_IsRunning(m_hnd) != X502_ERR_OK)) {
            uint8_t fbype;
            ok = X502_FlashRead(m_hnd, 0, &fbype, sizeof(fbype)) == X502_ERR_OK;
        }
        return ok;
    }

    bool x502::supportCustomOutFreq() const {
        return X502_CheckFeature(m_hnd, X502_FEATURE_OUT_FREQ_DIV) == X502_ERR_OK;
    }

    bool x502::adcIsRunning() const {
        return X502_IsRunning(m_hnd) == X502_ERR_OK ? true : false;
    }


    const x502Config &x502::devspecConfig() const {
        return static_cast<const x502Config &>(config());
    }

    const x502Info &x502::devspecInfo() const {
        return static_cast<const x502Info &>(devInfo());
    }

    void x502::getError(int err_code, LQError &err) {
        return getError(err_code, QString{}, err);
    }

    void x502::getError(int err_code, const QString &descr, LQError &err) {
        if (err_code != X502_ERR_OK) {
            err = error(err_code, descr);
        }
    }

    LQError x502::error(int err_code) {
        return error(err_code, QString{});
    }

    LQError LQMeas::x502::error(int err_code, const QString &descr) {
        return err_code == X502_ERR_OK ? LQError::Success() : x502NativeError::error(
                    err_code, (descr.isEmpty() ? descr : descr % QLatin1String(": "))
                    % QSTRING_FROM_CSTR(X502_GetErrorString(err_code)));
    }

    DeviceRef *x502::createRef() const {
        return new DeviceRef{devInfo(), iface()};
    }

    void x502::protOpen(OpenFlags flags, LQError &err) {
        Q_UNUSED(flags)

        t_x502_info devinfo;
        int32_t err_code {X502_OpenByDevRecord(m_hnd, &m_devrec)};
        if (err_code == X502_ERR_OK) {
            err_code = X502_GetDevInfo(m_hnd, &devinfo);
        }
        if (err_code == X502_ERR_OK) {
            m_opened = true;
            m_outStreams = 0;
            m_outRunning = false;
            m_inRunning = false;

            QScopedPointer<const x502Info> info(createInfo(devinfo));
            setDeviceInfo(*info);
        } else {
            X502_Close(m_hnd);
            getError(err_code, err);
        }
    }

    void x502::protClose(LQError &err) {
        getError(X502_Close(m_hnd), err);
        m_opened = false;
    }




    void x502::protConfigure(const DeviceConfig &cfg, LQError &err) {
        const x502Config *setCfg {qobject_cast<const x502Config *>(&cfg)};
        if (setCfg) {
            unsigned lch_cnt {0};
            QString errDescr;
            int err_code;


            if (!setCfg->bfIsEnabled()) {
                err_code = X502_SetMode(m_hnd, X502_MODE_FPGA);
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup FPGA mode");
                }
            } else {
                err_code = X502_BfLoadFirmware(m_hnd, QSTRING_TO_CSTR(setCfg->bfFirmwareFilename()));
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Load blackfin firmware");
                }
            }

            for (int ch=0; (ch < devTypeInfo().adc()->adcChannelsCnt()) && (err_code == X502_ERR_OK); ++ch) {
                if (setCfg->adc().adcChEnabled(ch)) {
                    x502AdcConfig::AdcChMode mode = setCfg->adc().adcChMode(ch);
                    uint32_t mode_code = mode == x502AdcConfig::AdcChMode::Comm ? X502_LCH_MODE_COMM :
                                        (mode == x502AdcConfig::AdcChMode::Zero ? X502_LCH_MODE_ZERO :
                                                                                  X502_LCH_MODE_DIFF);

                    err_code = X502_SetLChannel(m_hnd, lch_cnt, static_cast<uint32_t>(ch),
                                                 mode_code,
                                                 static_cast<uint32_t>(setCfg->adc().adcChRangeNum(ch)),
                                                 static_cast<uint32_t>(setCfg->adc().adcChAvg(ch)));
                    ++lch_cnt;
                    if (err_code != X502_ERR_OK) {
                        errDescr = tr("Setup channel %1 parameters").arg(QString::number(ch+1));
                    }
                }
            }

            if (err_code == X502_ERR_OK) {
                err_code = X502_SetLChannelCount(m_hnd, lch_cnt);
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup channels count");
                }
            }

            if (err_code == X502_ERR_OK) {
                err_code = X502_SetSyncMode(m_hnd, f_conv_sync_mode(setCfg->freqMode()));
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup sync mode");
                }
            }

            if (err_code == X502_ERR_OK) {
                err_code = X502_SetSyncStartMode(m_hnd, f_conv_sync_mode(setCfg->startMode()));
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup sync start mode");
                }
            }

            if (err_code == X502_ERR_OK) {
                err_code = X502_SetRefFreq(m_hnd, f_conv_ref_freq(setCfg->refFreqInternal()));
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup reference frequency");
                }
            }

            if (err_code == X502_ERR_OK) {
                err_code = X502_SetExtRefFreqValue(m_hnd, setCfg->refFreqExternal());
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup external reference frequency value");
                }
            }

            if (err_code == X502_ERR_OK) {
                double adc_freq = setCfg->adc().adcFreq();
                double frame_freq = setCfg->adc().adcChFreq();
                err_code = X502_SetAdcFreq(m_hnd, &adc_freq, &frame_freq);
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup adc frequency");
                }
            }

            if (err_code == X502_ERR_OK) {
                /* Изначально данная возможность не была реализована в прошивке
                   blackfin, не считаем этот случай ошибокой */
                int32_t err_code = X502_SetDigInPullup(m_hnd, setCfg->pullups());
                if (err_code == X502_ERR_NOT_IMPLEMENTED)
                    err_code = X502_ERR_OK;
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup pullup resistors");
                }
            }

            if (err_code == X502_ERR_OK) {
                double out_freq = setCfg->out().outSyncGenFreq();
                err_code = X502_SetOutFreq(m_hnd, &out_freq);
                if (err_code != X502_ERR_OK) {
                    errDescr = tr("Setup output frequency");
                }
            }

            if (err_code == X502_ERR_OK) {
                err_code = X502_Configure(m_hnd, 0);
            }

            if (err_code != X502_ERR_OK) {
                getError(err_code, errDescr, err);
            }
        } else {
            err = StdErrors::InvalidConfigType();
        }
    }

    void x502::protAdcEnable(LQError &err) {
        getError(X502_StreamsEnable(m_hnd, X502_STREAM_ADC), err);
    }

    void x502::protAdcDisable(LQError &err) {
        getError(X502_StreamsDisable(m_hnd, X502_STREAM_ADC), err);
    }

    void x502::protAdcStart(LQError &err) {
        m_receiver.setFrameSize(adcConfig()->adcEnabledChCnt());

        QMutexLocker lock{&m_streamLock};
        if (!m_outRunning) {
            getError(X502_StreamsStart(m_hnd), err);
           if (err.isSuccess())
                m_inRunning = true;
        }
    }

    void x502::protAdcStop(LQError &err) {
        QMutexLocker lock{&m_streamLock};
        m_inRunning = false;
        if (!m_outRunning) {
            getError(X502_StreamsStop(m_hnd), err);
        }
    }

    void x502::protAdcGetData(double *data, int size, unsigned flags,
                              unsigned tout, int &recvd_size, LQError &err) {
        unsigned recvd_words {0};
        uint32_t *wrds;

        err = m_receiver.getFrames(size, tout, &wrds, &recvd_words);


        if ((err.isSuccess()) && (recvd_words!=0)) {
            uint32_t proc_flags = X502_PROC_FLAGS_VOLT;
            uint32_t proc_wrds = recvd_words;

            getError(X502_ProcessAdcData(m_hnd, wrds, data, &proc_wrds, proc_flags), err);
            if (!err.isSuccess()) {
                LQMeasLog->error(tr("Process data error"), err, this);
            }

            recvd_words = proc_wrds;
        }

        if (err.isSuccess())
            recvd_size = static_cast<int>(recvd_words);
    }

    void x502::protInAsyncDig(unsigned &val, LQError &err) {
        getError(X502_AsyncInDig(m_hnd, &val), err);
    }

    void x502::rawWordsReceive(uint32_t *wrds, int size, unsigned tout,
                               int &recvd_size, LQError &err) {
        int32_t recv_res {X502_Recv(m_hnd, wrds, static_cast<uint32_t>(size), tout)};
        if (recv_res >= 0) {
            recvd_size = recv_res;
        } else {
            recvd_size = 0;
            getError(recv_res, err);
        }
    }

    void x502::setOutStreams(LQError &err) {
        int32_t err_code {X502_ERR_OK};
        uint32_t new_streams {0};
        if (devTypeInfo().out()->outDacChannelsCnt() >= X502_DAC_CH_CNT) {
            if (outSyncDacValidRamSignal(0))
                new_streams |= X502_STREAM_DAC1;
            if (outSyncDacValidRamSignal(1))
                new_streams |= X502_STREAM_DAC2;
        }
        bool outDig {false};
        for (int ch {0}; (ch < devTypeInfo().out()->outDigChannelsCnt()) && !outDig; ++ch) {
            if (outSyncDigValidRamSignal(ch))
                outDig = true;
        }
        if (outDig)
            new_streams |= X502_STREAM_DOUT;

        uint32_t cur_stream;
        /* Явно получаем список разрешенных потоков, а не используем из
           m_outStreams, чтобы избежать возможной раcсинхронизации в случае
           работы с BlackFin */
        err_code = X502_GetEnabledStreams(m_hnd, &cur_stream);
        if (err_code == X502_ERR_OK) {
            cur_stream &= X502_STREAM_ALL_OUT;

            uint32_t streams_en = new_streams & ~cur_stream;
            uint32_t streams_dis = ~new_streams & cur_stream;

            if (streams_en) {
                err_code = X502_StreamsEnable(m_hnd, streams_en);
            }
            if (streams_dis && (err_code == X502_ERR_OK)) {
                err_code = X502_StreamsDisable(m_hnd, streams_dis);
            }
        }

        if (err_code == X502_ERR_OK) {
            m_outStreams = new_streams;
        } else {
            getError(err_code, err);
        }
    }

    void x502::protOutAsyncDac(int ch, double val, LQError &err) {
        getError(X502_AsyncOutDac(m_hnd, static_cast<uint32_t>(ch), val,
                                  X502_DAC_FLAGS_VOLT | X502_DAC_FLAGS_CALIBR), err);
    }

    void x502::protOutAsyncDig(unsigned val, unsigned mask, LQError &err) {
        return getError(X502_AsyncOutDig(m_hnd, val, mask), err);
    }

    void x502::rawWordsSend(const uint32_t *wrds, int size, unsigned tout,
                            int &sent_size, LQError &err) {

        int send_res = X502_Send(m_hnd, wrds, static_cast<uint32_t>(size), tout);
        if (send_res >= 0) {
            sent_size = send_res;
        } else {
            sent_size = 0;
            getError(send_res, err);
        }
    }

    const DeviceInterface &x502::devrecIface(const t_x502_devrec &devrec) {
        return devrec.iface == X502_IFACE_PCI ? DeviceInterfaces::pcie() :
               devrec.iface == X502_IFACE_USB ? DeviceInterfaces::usb() :
               devrec.iface == X502_IFACE_ETH ? DeviceInterfaces::ethernet() :
                                                DeviceInterfaces::unknown();
    }

    void x502::privOutStreamInit(LQError &err) {
        return setOutStreams(err);
    }

    void x502::privOutStreamStartRequest(LQError &err) {
        QMutexLocker lock(&m_streamLock);
        if (!m_inRunning)
            getError(X502_StreamsStart(m_hnd), err);
        if (err.isSuccess())
            m_outRunning = true;
    }

    void x502::privOutStreamStop(unsigned tout, LQError &err) {
        if (m_sender.unsentWordsCnt()) {
            m_sender.flushData(tout, err);
            m_sender.clearData();
        }

        QMutexLocker lock(&m_streamLock);
        m_outRunning = false;
        if (!m_inRunning) {
            getError(X502_StreamsStop(m_hnd), err);
        }
    }

    void x502::privOutSyncSendData(const double *dac_data, int dac_size,
                                    const unsigned *dig_data, int dig_size,
                                    unsigned flags, unsigned tout, LQError &err) {
        Q_UNUSED(flags)
        if (tout == 0) {
            tout = 3000 + static_cast<unsigned>((dac_size + dig_size)/1000);
        }

        if (m_sender.unsentWordsCnt()) {
            m_sender.flushData(tout, err);
        }

        if (err.isSuccess()) {
            if (m_sender.unsentWordsCnt()!=0) {
                err = StdErrors::SendBusy();
            } else {
                int snd_dac_req_size = dac_size;
                int snd_dig_req_size = dig_size;
                int snd_size = snd_dac_req_size  + snd_dig_req_size;
                QScopedArrayPointer<uint32_t> wrds (new uint32_t[static_cast<size_t>(snd_size)]);

                if ((m_outStreams & X502_STREAM_DAC1) && (m_outStreams & X502_STREAM_DAC2)) {
                    dac_size/=2;
                    QScopedArrayPointer<double> dac1 (new double[static_cast<size_t>(dac_size)]);
                    QScopedArrayPointer<double> dac2 (new double[static_cast<size_t>(dac_size)]);
                    for (int i=0; i < static_cast<int>(dac_size); ++i) {
                        dac1[i] = *dac_data++;
                        dac2[i] = *dac_data++;
                    }
                    getError(X502_PrepareData(m_hnd, dac1.data(), dac2.data(),
                                              reinterpret_cast<const uint32_t*>(dig_data),
                                              static_cast<uint32_t>(dac_size),
                                              X502_DAC_FLAGS_CALIBR  | X502_DAC_FLAGS_VOLT,
                                              wrds.data()), err);
                } else {
                    getError(X502_PrepareData(m_hnd, dac_data, dac_data,
                                              reinterpret_cast<const uint32_t*>(dig_data),
                                              static_cast<uint32_t>(dig_size),
                                              X502_DAC_FLAGS_CALIBR  | X502_DAC_FLAGS_VOLT,
                                              wrds.data()), err);
                }

                if (!err.isSuccess()) {
                    LQMeasLog->error(tr("Prepare data error"), err, this);
                } else {
                    m_sender.setFrameSize(snd_size);
                    m_sender.putFrames(wrds.data(), snd_size, tout, nullptr, err);
                }
            }
        }
    }

    bool x502::privOutSyncHasDefferedData() const {
        return m_sender.unsentWordsCnt() > 0;
    }

    void x502::privOutSyncTryFlushData(unsigned tout, LQError &err) {
        m_sender.flushData(tout, err);
    }

    void x502::privOutSyncSendFinish(LQError &err) {
        Q_UNUSED(err)
        m_sender.setFrameSize(0);
    }

    bool x502::privOutSyncUnderflowOccured() const {
        bool ret = false;
        uint32_t status;
        int32_t res = X502_OutGetStatusFlags(m_hnd, &status);
        if (res == X502_ERR_OK) {
            if (status & X502_OUT_STATUS_FLAG_BUF_WAS_EMPTY)
                ret = true;
        }
        return ret;
    }

    void x502::privOutCycleLoadStart(const OutRamSignalGenerator &generator, int size, LQError &err) {
        setOutStreams(err);
        if (err.isSuccess()) {
            int chCnt = generator.dacSignalsCount() + (generator.digSignalsCount() == 0 ? 0 : 1);
            getError(X502_OutCycleLoadStart(m_hnd, static_cast<uint32_t>(size*chCnt)), err);
        }
    }

    void x502::privOutCycleLoadFinish(LQError &err) {
        return getError(X502_OutCycleSetup(m_hnd, !m_outRunning ? X502_OUT_CYCLE_FLAGS_WAIT_DONE : 0), err);
    }

    void x502::privOutCycleGenStartRequest(LQError &err) {
        QMutexLocker lock{&m_streamLock};
        if (!m_inRunning && !m_outRunning)
            getError(X502_StreamsStart(m_hnd), err);
        if (err.isSuccess())
            m_outRunning = true;
    }

    void x502::privOutCycleGenUpdateRequest(LQError &err) {
        privOutCycleGenStartRequest(err);
    }

    void x502::privOutCycleGenStopRequest(unsigned tout, LQError &err) {
        getError(X502_OutCycleStop(m_hnd, tout), err);
    }

    void x502::privOutCycleGenStop(LQError &err) {
        QMutexLocker lock{&m_streamLock};
        m_outRunning = false;
        /** @note может есть смысл делать X502_OutCycleCheckSetupDone() */
        if (!m_inRunning) {
            /* останавливаем поток только если поток ввода тоже уже остановлен */
            getError(X502_StreamsStop(m_hnd), err);
        }
    }



}
