#include "x502Config.h"
#include "x502TypeInfo.h"
#include "x502.h"
#include "lqmeas/EnumConfigKey.h"
#include "lqmeas/StdConfigKeys.h"
#include "lqmeas/lqtdefs.h"
#include <QJsonObject>
#include <QJsonArray>

namespace LQMeas {
    static const QLatin1String cfgkey_sync_freq_mode         {"SyncFreqMode"};
    static const QLatin1String cfgkey_sync_start_mode        {StdConfigKeys::syncStartMode()};
    static const QLatin1String cfgkey_int_ref_freq           {"IntRefFreq"};
    static const QLatin1String cfgkey_ext_ref_freq           {"ExtRefFreq"};
    static const QLatin1String cfgkey_pullups                {StdConfigKeys::pullups()};
    static const QLatin1String cfgkey_bf_obj                 {"Blackfin"};
    static const QLatin1String cfgkey_bf_enabled             {StdConfigKeys::enabled()};
    static const QLatin1String cfgkey_bf_fw_filename         {"FirmwareFile"};

    typedef EnumConfigKey<x502Config::SyncMode> X502SyncModeKey;
    static const X502SyncModeKey sync_modes[] = {
        {QStringLiteral("int"),         x502Config::SyncMode::Internal},
        {QStringLiteral("extmstr"),     x502Config::SyncMode::ExternalMaster},
        {QStringLiteral("syn1rise"),    x502Config::SyncMode::SYN1Rise},
        {QStringLiteral("syn1fall"),    x502Config::SyncMode::SYN1Fall},
        {QStringLiteral("syn2rise"),    x502Config::SyncMode::SYN2Rise},
        {QStringLiteral("syn2fall"),    x502Config::SyncMode::SYN2Fall},
        {X502SyncModeKey::defaultKey(), x502Config::SyncMode::Internal},
    };
    typedef EnumConfigKey<x502Config::RefFreq> X502IntRefFreqKey;
    static const X502IntRefFreqKey ref_freqs[] = {
        {QStringLiteral("2000K"),       x502Config::RefFreq::Freq_2000KHz},
        {QStringLiteral("1500K"),       x502Config::RefFreq::Freq_1500KHz},
        {X502IntRefFreqKey::defaultKey(), x502Config::RefFreq::Freq_2000KHz},
    };


    const QString &x502Config::typeConfigName() {
        static const QString str {QStringLiteral("x502")};
        return str;
    }

    x502Config::x502Config() :
        m_adc{this, x502TypeInfo::defaultInfo()},
        m_out{this, x502TypeInfo::defaultInfo()} {

        memset(&params.Sync, 0, sizeof(params.Sync));
        params.Pullups = 0;
        params.Blackfin.Enabled = false;


        results.DinFreq = adc().adcInfo()->adcFreqMax();
        params.Sync.FreqMode = params.Sync.StartMode = SyncMode::Internal;
        params.Sync.RefFreqInternal = RefFreq::Freq_2000KHz;


        init();
    }

    x502Config::x502Config(const x502Config &cfg) :
        m_adc{this, cfg.adc()},
        m_out{this, cfg.out()} {

        memcpy(&params.Sync, &cfg.params.Sync, sizeof(params.Sync));
        memcpy(&results, &cfg.results, sizeof(results));
        params.Pullups = cfg.params.Pullups;
        params.Blackfin.Enabled = cfg.params.Blackfin.Enabled;
        params.Blackfin.FileName = cfg.params.Blackfin.FileName;

        init();
    }



    double x502Config::externalRefFreqMax() {
        return  X502_EXT_REF_FREQ_MAX;
    }

    void x502Config::setRefFreqExternal(double val) {
        if (val > externalRefFreqMax())
            val = externalRefFreqMax();
        if (!LQREAL_IS_EQUAL(val, params.Sync.RefFreqExternal)) {
            params.Sync.RefFreqExternal = val;
            notifyConfigChanged();
        }
    }

    void x502Config::setBfFirmwareFilename(QString filename) {
        if (params.Blackfin.FileName != filename) {
            params.Blackfin.FileName = filename;
            notifyConfigChanged();
        }
    }

    double x502Config::refFreq() const {
        return params.Sync.FreqMode == SyncMode::Internal ?
                    typeInternalRefFreqVal(params.Sync.RefFreqInternal) : params.Sync.RefFreqExternal;
    }


    void x502Config::setPullups(uint32_t pullups_flags) {
        if (params.Pullups != pullups_flags) {
            params.Pullups = pullups_flags;
            notifyConfigChanged();
        }
    }

    void x502Config::setBfEnalbed(bool en) {
        if (params.Blackfin.Enabled != en) {
            params.Blackfin.Enabled = en;
            notifyConfigChanged();
        }
    }

    void x502Config::setFreqMode(x502Config::SyncMode mode) {
        if (params.Sync.FreqMode != mode) {
            params.Sync.FreqMode = mode;
            notifyConfigChanged();
        }
    }

    void x502Config::setStartMode(x502Config::SyncMode mode) {
        if (params.Sync.StartMode != mode) {
            params.Sync.StartMode = mode;
            notifyConfigChanged();
        }
    }

    void x502Config::setRefFreqInternal(x502Config::RefFreq val) {
        if (params.Sync.RefFreqInternal != val) {
            params.Sync.RefFreqInternal = val;
            notifyConfigChanged();
        }
    }

    void x502Config::protSave(QJsonObject &cfgObj) const {
        cfgObj[cfgkey_sync_freq_mode]  = X502SyncModeKey::getKey(sync_modes, params.Sync.FreqMode);
        cfgObj[cfgkey_sync_start_mode] = X502SyncModeKey::getKey(sync_modes, params.Sync.StartMode);
        cfgObj[cfgkey_int_ref_freq]    = X502IntRefFreqKey::getKey(ref_freqs, params.Sync.RefFreqInternal);
        cfgObj[cfgkey_ext_ref_freq]    = params.Sync.RefFreqExternal;

        cfgObj[cfgkey_pullups] =  static_cast<int>(params.Pullups);

        QJsonObject bfObj;
        bfObj[cfgkey_bf_enabled]   = params.Blackfin.Enabled;
        bfObj[cfgkey_bf_fw_filename] = params.Blackfin.FileName;
        cfgObj[cfgkey_bf_obj] = bfObj;
    }

    void x502Config::protLoad(const QJsonObject &cfgObj) {
        params.Sync.FreqMode = X502SyncModeKey::getValue(sync_modes, cfgObj[cfgkey_sync_freq_mode].toString());
        params.Sync.StartMode = X502SyncModeKey::getValue(sync_modes, cfgObj[cfgkey_sync_start_mode].toString());
        params.Sync.RefFreqInternal = X502IntRefFreqKey::getValue(ref_freqs, cfgObj[cfgkey_int_ref_freq].toString());
        double ext_ref_freq = cfgObj[cfgkey_ext_ref_freq].toDouble(externalRefFreqMax());

        if ((ext_ref_freq <= 0) || (ext_ref_freq > externalRefFreqMax()))
            ext_ref_freq = externalRefFreqMax();
        params.Sync.RefFreqExternal = ext_ref_freq;
        params.Pullups = static_cast<uint32_t>(cfgObj[cfgkey_pullups].toInt());

        const QJsonObject &bfObj = cfgObj[cfgkey_bf_obj].toObject();
        params.Blackfin.Enabled = bfObj[cfgkey_bf_enabled].toBool(false);
        params.Blackfin.FileName = bfObj[cfgkey_bf_fw_filename].toString();
    }

    void x502Config::protUpdate() {
        double ref_freq = refFreq();

        results.DinFreq = ref_freq;


        X502_CalcDinFreq(ref_freq, &results.DinFreq, nullptr);
    }


}
