#include "x502Resolver.h"
#include "lqmeas/lqtdefs.h"
#include "lqmeas_config.h"
#include <QScopedArrayPointer>
#include "x502.h"
#include "lqmeas/devs/DeviceRef.h"
#include "lqmeas/devs/DeviceInterface.h"

#ifdef LQMEAS_DEV_E502
    #include "e502api.h"
    #include "modules/E502/E502.h"
    #include "modules/E502/E502TypeInfo.h"
    #include "modules/E502/E502DeviceRefIPAddr.h"
    #include "modules/E502/netBrowser/E502NetSvcRecord.h"
    #include "modules/E502/netBrowser/E502DeviceRefNetSvc.h"
#endif
#ifdef LQMEAS_DEV_L502
    #include "l502api.h"
    #include "modules/L502/L502.h"
    #include "modules/L502/L502TypeInfo.h"
#endif

namespace LQMeas {
    x502Resolver::x502Resolver() {
    }


    QList<QSharedPointer<Device> > x502Resolver::protGetDevList(
            const QList<const DeviceRef *> &devRefs, ResolveFlags flags) {
        Q_UNUSED(flags)

        QList<QSharedPointer<Device> > newDevices;
        QList<QSharedPointer<Device> > oldDevices {m_devs};

        /* поиск автоопределяемых устройств, подключенных по PCIe или USB */
        {
            uint32_t devcnt_usb {0};
            uint32_t devcnt_pci {0};
            uint32_t devcnt {0};
    #ifdef LQMEAS_DEV_E502
            E502_UsbGetDevRecordsList(nullptr, 0, 0, &devcnt_usb);
    #endif
    #ifdef LQMEAS_DEV_L502
            L502_GetDevRecordsList(nullptr, 0, 0, &devcnt_pci);
    #endif
            devcnt = devcnt_usb + devcnt_pci;

            if (devcnt > 0) {
                 int fnd_devcnt {0};
                int32_t res = X502_ERR_OK;
                QScopedArrayPointer<t_x502_devrec> devrecs{new t_x502_devrec[devcnt]};
                /* получаем информацию по всем устройствам  */
    #ifdef LQMEAS_DEV_E502
                res = E502_UsbGetDevRecordsList(devrecs.data(), devcnt_usb, 0, nullptr);
                if (res >= 0) {
                    fnd_devcnt += res;
                }
    #endif
    #ifdef LQMEAS_DEV_L502
                res = L502_GetDevRecordsList(&devrecs.data()[fnd_devcnt], devcnt_pci, 0, nullptr);
                if (res >= 0) {
                    fnd_devcnt += res;
                }
    #endif
                if (fnd_devcnt > 0) {
                    for (int devrec_idx = 0; devrec_idx < fnd_devcnt; ++devrec_idx) {
                        t_x502_devrec *prec {&devrecs[devrec_idx]};
                        QString devserial {QSTRING_FROM_CSTR(prec->serial)};
                        QString devname {QSTRING_FROM_CSTR(prec->devname)};
                        QSharedPointer<Device> fndDev;
                        if (devname == QLatin1String("E502")) {
                            devname =  E502TypeInfo::name();
                        } else if (devname == QLatin1String("L502")) {
                            devname = L502TypeInfo::name();
                        }

                        for (const QSharedPointer<Device> &dev : oldDevices) {
                            if ((dev->serial() == devserial) &&
                                    (dev->name() == devname) &&
                                    (dev->iface() != DeviceInterfaces::ethernet())) {
                                fndDev = dev;
                                break;
                            }
                        }

                        if (fndDev) {
                            oldDevices.removeOne(fndDev);
                            newDevices.append(fndDev);

                            QSharedPointer<x502> x502dev {qSharedPointerObjectCast<x502>(fndDev)};
                            x502dev->setNewDevRecord(*prec);
                        } else {
    #ifdef LQMEAS_DEV_E502
                            QSharedPointer<Device> newDev;
                            if (devname == E502TypeInfo::name()) {
                                newDev = QSharedPointer<Device>(new E502{*prec});
                            }
    #endif
    #ifdef LQMEAS_DEV_L502
                            if (devname == L502TypeInfo::name()) {
                                newDev = QSharedPointer<Device>(new L502{*prec});
                            }
    #endif
                            if (newDev) {
                                newDevices.append(newDev);
                            } else {
                                X502_FreeDevRecordList(prec,1);
                            }
                        }
                    }
                }
            }
        }

    #ifdef LQMEAS_DEV_E502
        QList<QSharedPointer<DeviceNetSvcRecord> > netRecs {m_netBrowser.records()};

        for (const QSharedPointer<DeviceNetSvcRecord> &netRec : netRecs) {
            const QSharedPointer<const E502NetSvcRecord> e502rec {qSharedPointerCast<const E502NetSvcRecord>(netRec)};
            QSharedPointer<Device> fndDev;
            for (const QSharedPointer<Device> &dev : oldDevices) {
                E502 *e502 {qobject_cast<E502*>(dev.data())};
                if (e502 && e502->netRecord() && netRec->isEqual(*e502->netRecord())) {
                    fndDev = dev;
                    break;
                }
            }

            if (fndDev) {
                oldDevices.removeOne(fndDev);
            } else {
                /* если есть ссылка на это устройство, то берем параметры подключения
                 * из ссылки (ссылка может переназначать таймаут подключения) */
                const E502DeviceRefNetSvc *fndRef = nullptr;
                for (const DeviceRef *ref : devRefs) {
                    if (ref->refTypeID() == E502DeviceRefNetSvc::typeRefTypeID()) {
                        const E502DeviceRefNetSvc *svcRef {static_cast<const E502DeviceRefNetSvc *>(ref)};
                        if (svcRef->checkNetSvcRecord(*e502rec)) {
                            fndRef = svcRef;
                            break;
                        }
                    }
                }
                fndDev = QSharedPointer<Device>{fndRef ? new E502{e502rec, *fndRef}
                                                       : new E502{e502rec, E502::defaultEthConnectionTout()}};
            }

            if (fndDev)
                newDevices.append(fndDev);
        }
    #endif

        for (const DeviceRef *ref : devRefs) {
            if (!ref->isAutoResolved()) {
    #ifdef LQMEAS_DEV_E502
                if (ref->refTypeID() == E502DeviceRefIPAddr::typeRefTypeID()) {
                    const E502DeviceRefIPAddr *ipRef {static_cast<const E502DeviceRefIPAddr *>(ref)};
                    QSharedPointer<Device> fndDev;
                    for (const QSharedPointer<Device> &dev : oldDevices) {
                        if (ref->checkDevice(*dev) != DeviceRef::DevCheckResult::BadDevice) {
                            fndDev = dev;
                            break;
                        }
                    }

                    if (fndDev) {
                        oldDevices.removeOne(fndDev);
                    } else {
                        /* проверяем на всякий случай среди новых устройств, что это не
                         * повторная ссылка */
                        bool fnd_newDev {false};
                        if (!fndDev) {
                            for (const QSharedPointer<Device> &dev : newDevices) {
                                if (ref->checkDevice(*dev) != DeviceRef::DevCheckResult::BadDevice) {
                                    fnd_newDev = true;
                                    break;
                                }
                            }
                        }

                        if (!fndDev && !fnd_newDev)
                            fndDev = QSharedPointer<Device>{new E502{*ipRef}};
                    }

                    if (fndDev)
                        newDevices.append(fndDev);
                }
    #endif
            }
        }

#ifdef Q_OS_WIN
        /* т.к. для E502 по крайней мере для Windows мы не можем определить наличие
         * устройства E502, подключенного по USB, если оно уже открыто, то все открытые
         * устройства добавляем в список присутствующих */
        for(const QSharedPointer<Device> &oldDev : oldDevices) {
            if (oldDev->isOpened() && (oldDev->name() == E502TypeInfo::name())) {
                newDevices.append(oldDev);
            }
        }
#endif

        m_devs = newDevices;

        return newDevices;
    }

    x502Resolver &x502Resolver::resolver() {
        static x502Resolver resolver;
        return resolver;
    }
}
